
/* special purpose post-optimizer for cc65 */

#include <stdio.h>
#include <string.h>

int ln = 0;

int matchstr(line, str)
char * line;
char * str;
{
  for ( ; *str ; )
    {
      if (*line != *str)
	return(0);
      ++line;
      ++str;
    }
  return(1);
}

int matchjsr(line, str)
char *line; char *str;
{
  return(matchstr(line, "\tjsr\t") && (matchstr(line+5, str)));
}

/* returns ptr to stuff after ldy if it's usable, ie constant 0..7 */
char * matchldy(line)
char * line;
{
  if (matchstr(line, "\tldy\t#"))
    {
      if ((line[6] >= '0') && (line[6] < '8') && (line[7] == '\n'))
	return(line + 6);
    }
  return(0);
}

/* routines that can use implied offsets */
char * impl[] = 
{
  "ldaxysp",
  "ldaysp",
  "staxysp",
  "staysp",
  "pushwysp",
  "pushbysp",
  0
  };

/* handed 2 lines, return ptr to one to new one if optimized */

/* change : lda L... or lda _...
            ina or dea
            sta L...
   into   : inc L...

BUT only if a branch or another lda or ldx follows
*/
char * opt_dea(char *l1,char *l2,char *l3,char *l4)
{

  if (!(matchstr(l1,"\tlda\tL") || matchstr(l1,"\tlda\t_"))) return 0;
  if (!(matchstr(l2,"\tdea") || matchstr(l2,"\tina"))) return 0;
  if (!matchstr(l3,"\tsta")) return 0;
  if (matchstr(l4,"\tlb") || matchstr(l4,"\tld") ||
      matchjsr(l4,"_") || l4[0]=='L')
  { 
    l1[1]=l2[1]; l1[2]=l2[2]; l1[3]='c';
    return l1;
  }else
    return 0;
}

/* delete a jsr tstax if it follows a runtime-operation */
int opt_tstax(char *l,char *l1)
{
  return ( (matchstr(l+8,"to")!= NULL) && matchjsr(l1,"tstax"));
}


/* change : lda #0
            sta L... or sta _...
   into   : stz L...
*/

char * opt_stz(char *l1,char *l2)
{  
  if ( !matchstr(l1,"\tlda \t#0") ) return 0;
                       /* ^ it was ldax #0 !! */
 
  if ( !(matchstr(l2,"\tsta\tL") || matchstr(l2,"\tsta\t_")) ) return 0;

  l2[3] = 'z';
  return (l2);
}

int readline(f, b)
FILE * f;
char * b;
{
  int c;
  char * p;
  
  ++ln;
  for (p = b ; ((c = fgetc(f)) != EOF) ; )
    {
      *p++ = c;
      if (c == '\n')
	break;
    }
  *p = '\0';
  return(p != b);
}

char lbuf1[80];
char lbuf2[80];
char lbuf3[80];
char lbuf4[80];

main(argc, argv)
int argc;
char ** argv;
{
  FILE * inf, * outf;
  char * line;
  char *l1,*l2,*l3,*l4;
  int opt = 0;
  
  if (!(inf = fopen(argv[1], "r")))
    {
      printf("Can't open %s", argv[1]);
      exit(1);
    }
  outf = fopen("xopt.tmp", "w");

  l1 = lbuf1;
  l2 = lbuf2;
  l3 = lbuf3;
  l4 = lbuf4;
  
  readline(inf, l1);
  readline(inf, l2);
  readline(inf, l3);
  while(readline(inf, l4))
    {
      if ( line = opt_dea(l1, l2, l3, l4) )
      {
	++opt;
	fputs(line,outf);
        line = l1;
        l1 = l4;
        l4 = line;
	if ( !readline(inf,l2) || !readline(inf,l3)) break;
        
      }
      else if ( opt_tstax(l1,l2) )
      {
	++opt;
	fputs(l1,outf);
	line = l1;
	l1 = l3;
	l3 = l2;
	l2 = l3;
	l4 = line;
	if (!readline(inf,l3)) break;
     }else if ( line = opt_stz(l1,l2) )
     {
        ++opt;
        fputs(line,outf);
	line = l1;
	l1 = l3;
        l3 = l2;
	l2 = l4;
	l4 = line;
	if (!readline(inf,l3)) break;
     }else	
     {
	fputs(l1,outf);
	line = l1;
	l1 = l2;
	l2 = l3;
        l3 = l4;
	l4 = line;
     }
    }
  fputs(l1,outf);
  fputs(l2,outf);
  fputs(l3,outf);
  fclose(inf);
  fclose(outf);
  unlink(argv[1]);
  rename("xopt.tmp", argv[1]);
  return 0;
}
